/*
 * This file is part of hildon-input-method-plugins-sdk
 *
 * Copyright (C) 2006 Nokia Corporation. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#include <string.h>
#include <glib.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gtk/gtkwidget.h>
#include "hildon-im-plugin.h"
#include "hildon-im-keyboard.h"

#define HIMSDK_VKB_TYPE himsdk_vkb_get_type()
#define HIMSDK_VKB(obj) GTK_CHECK_CAST(obj, himsdk_vkb_get_type(), HimsdkVKB)
#define HIMSDK_VKB_CLASS(klass) \
        GTK_CHECK_CLASS_CAST(klass, himsdk_vkb_get_type, \
                             HimsdkVKBClass)
#define IS_HIMSDK_VKB(obj) \
        GTK_CHECK_TYPE(obj, himsdk_vkb_get_type())
#define HIMSDK_VKB_GET_PRIVATE(obj) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((obj), HIMSDK_VKB_TYPE,\
                                      HimsdkVKBPrivate))

#define FULLSCREEN_WIDTH   600
#define START_X             20
#define START_Y              0
#define HEIGHT              35
#define MAX_KEYPAD_KEYS      3
#define KEYBOARD_ROWS        4
#define KEYPAD_ROWS          4
#define KEYPAD_MARGIN_RIGHT 18

#define KEY_LABEL_MARGIN     3

#define TOTAL_KEYPAD_KEYS   MAX_KEYPAD_KEYS * KEYPAD_ROWS

enum {
  SCREEN_NORMAL,
  SCREEN_FULL,
  SCREEN_MAX
};

enum {
  CASE_UPPER,
  CASE_LOWER,
  CASE_MAX
};

enum {
  KEYSIZE_NORMAL,
  KEYSIZE_BIG,
  KEYSIZE_SPACE,

  KEYSIZE_MAX
};

static const gint button_sizes [SCREEN_MAX][KEYSIZE_MAX] = {
  { 34, 51, -1 },
  { 40, 60, -1 }
};

static const gint keyboard_keys [KEYBOARD_ROWS] = {
  11, 12, 11, 1
};

static const gchar *keyboard_layout [CASE_MAX][KEYBOARD_ROWS] = {
  { "QWERTYUIOP^", "ASDFGHJKL:#|", "ZXCVBNM<>\\~", " " },
  { "qwertyuiop@", "asdfghjkl;'!", "zxcvbnm,./?", " " }
};

static const gchar *keypad_layout [CASE_MAX][KEYPAD_ROWS] = {
  { "'\"Â£", "$â¬%", "&()", "_*+" },
  { "123", "456", "789", "-0=" }
};

typedef struct _vkbkey VKBKey;

static GType himsdk_vkb_type = 0;
static GtkWidgetClass *parent_class = NULL;

typedef struct {
  GtkWidgetClass parent;
} HimsdkVKBClass;

typedef struct {
  GtkWidget parent;
} HimsdkVKB;

typedef struct {
  HildonIMKeyboard  *keyboard;
  GtkStyle          *style;
  VKBKey            *keys, 
                    *last_key;
  PangoLayout       *layout;

  gint               screen_mode;
  gint               case_mode;
  gint               total_keys;
  
  gint               keypad_start_x;
} HimsdkVKBPrivate;

struct _vkbkey {
  gchar             *upper;
  gint               upper_length;
  gchar             *lower;
  gint               lower_length;
  gint               x, y, width, height;
};

GType himsdk_vkb_get_type (void);
GtkWidget *himsdk_vkb_new (HildonIMKeyboard *kbd);

/* Prototype declarations */
static void himsdk_vkb_iface_init (HildonIMPluginIface *iface);
static void himsdk_vkb_class_init (HimsdkVKBClass *klass);
static void finalize(GObject *obj);
static void get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void himsdk_vkb_init (HimsdkVKB *vkb);
static void enable (HildonIMPlugin *plugin, gboolean init);
static void disable(HildonIMPlugin *plugin);
static void input_mode_changed (HildonIMPlugin *plugin);
static void clear(HildonIMPlugin *plugin);
static void client_widget_changed (HildonIMPlugin *plugin);
static void save_data(HildonIMPlugin *plugin);
static void mode_a(HildonIMPlugin *plugin); 
static void mode_b(HildonIMPlugin *plugin);
static void language (HildonIMPlugin *plugin);
static void language_settings_changed (HildonIMPlugin *plugin,
    gint index, gint value);
static void backspace (HildonIMPlugin *plugin);
static void enter (HildonIMPlugin *plugin);
static void tab (HildonIMPlugin *plugin);
static void settings_changed (HildonIMPlugin *plugin, const gchar *key, const GConfValue *value);

static void draw_keyboard (HimsdkVKB *);
static void update_layout (HimsdkVKB *);
static void send_key (HimsdkVKB *, VKBKey *key);

static gboolean expose_cb (GtkWidget *, GdkEventExpose *);
static gboolean destroy_cb (GtkWidget *, GdkEventAny *);
static void realize_cb (GtkWidget *);
static void size_request_cb (GtkWidget *, GtkRequisition *);
static void size_allocate_cb (GtkWidget *, GtkAllocation *);
static gboolean button_press_cb (GtkWidget *, GdkEventButton *);
static gboolean button_release_cb (GtkWidget *, GdkEventButton *);

/**
 * Module functions
 *
 **/

HildonIMPlugin* 
module_create (HildonIMKeyboard *keyboard)
{
  return HILDON_IM_PLUGIN (himsdk_vkb_new (keyboard));
}

void
module_exit(void)
{
  /* empty */
}

void
module_init(GTypeModule *module)
{
  static const GTypeInfo type_info = {
    sizeof(HimsdkVKBClass),
    NULL, /* base_init */
    NULL, /* base_finalize */
    (GClassInitFunc) himsdk_vkb_class_init,
    NULL, /* class_finalize */
    NULL, /* class_data */
    sizeof(HimsdkVKB),
    0, /* n_preallocs */
    (GInstanceInitFunc) himsdk_vkb_init,
  };
  
  static const GInterfaceInfo plugin_info = {
    (GInterfaceInitFunc) himsdk_vkb_iface_init,
    NULL, /* interface_finalize */
    NULL, /* interface_data */
  };

  himsdk_vkb_type =
          g_type_module_register_type(module,
                                      GTK_TYPE_WIDGET, "HimsdkVKB",
                                      &type_info,
                                      0);
  
  g_type_module_add_interface(module,
                              HIMSDK_VKB_TYPE,
                              HILDON_IM_TYPE_PLUGIN,
                              &plugin_info);
}

GType
himsdk_vkb_get_type (void)
{
  return himsdk_vkb_type;
}

/* Standard GTK stuff */
static void
himsdk_vkb_iface_init (HildonIMPluginIface *iface)
{
  iface->enable = enable;
  iface->disable = disable;
  iface->enter = enter;
  iface->tab = tab;
  iface->backspace = backspace;
  iface->clear = clear;
  iface->input_mode_changed = input_mode_changed;
  iface->client_widget_changed = client_widget_changed;
  iface->save_data = save_data;
  iface->language = language;
  iface->mode_a = mode_a;
  iface->mode_b = mode_b;
  iface->language_settings_changed = language_settings_changed;
  iface->settings_changed = settings_changed;

  return;
}

static void
himsdk_vkb_class_init (HimsdkVKBClass *klass)
{
  GObjectClass *object_class;
  GtkObjectClass *gtk_object_class;
  GtkWidgetClass *widget_class;

  parent_class = g_type_class_peek_parent(klass);
  g_type_class_add_private(klass, sizeof(HimsdkVKBPrivate));

  object_class = G_OBJECT_CLASS(klass);
  gtk_object_class = GTK_OBJECT_CLASS(klass);
  widget_class = GTK_WIDGET_CLASS(klass);

  object_class->set_property  = set_property;
  object_class->get_property  = get_property;
  object_class->finalize      = finalize;

  widget_class->realize       = realize_cb;
  widget_class->destroy_event = destroy_cb;
  widget_class->expose_event  = expose_cb;
  widget_class->size_request  = size_request_cb;
  widget_class->size_allocate = size_allocate_cb;

  widget_class->button_press_event    = button_press_cb;
  widget_class->button_release_event  = button_release_cb;

  g_object_class_install_property(object_class, PROP_KEYBOARD,
                                  g_param_spec_object("keyboard", 
                                  "Keyboard",
                                  "Keyboard that uses plugin",
                                  HILDON_IM_TYPE_KEYBOARD,
                                  G_PARAM_READWRITE
                                  | G_PARAM_CONSTRUCT_ONLY));  
}

/* Input Method plugin information.
 * This structure tells the main UI about this plugin */
const hildon_im_plugin_info *
hildon_im_plugin_get_info(void)
{
  static const hildon_im_plugin_info info =
  {
    "himsdk_vkb_new",                        /* new func */
    "himsdk_vkb", /* name */
    "Keyboard (SDK)", /* menu title */
    NULL, /* gettext */
    TRUE, /* visible in menu */
    FALSE, /* cached */
    IM_TYPE_KEYBOARD,                                    /* type */
    IM_GROUP_WESTERN,                                   /* group */
    DEFAULT_PRIORITY,                                   /* priority */
    NULL, /* no other plugin in SCV button */
    ""      /* help page */
  };

  return &info;
}

/*
 * This function returns the list of available languages supported
 * by the plugin.
 * */
gchar ** 
hildon_im_plugin_get_available_languages (gboolean *free)
{
  static gchar *list[] = { "en_GB", NULL };
  *free = FALSE;

  return list;
}

static void 
finalize(GObject *obj)
{
  if (G_OBJECT_CLASS(parent_class)->finalize)
  {
    G_OBJECT_CLASS(parent_class)->finalize(obj);
  }
}

static void
get_property (GObject *object,
    guint prop_id,
    GValue *value,
    GParamSpec *pspec)
{
  HimsdkVKBPrivate *priv;

  g_return_if_fail (IS_HIMSDK_VKB(object));
  priv = HIMSDK_VKB_GET_PRIVATE(object);

  switch (prop_id)
  {
    case PROP_KEYBOARD:
      g_value_set_object(value, priv->keyboard);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
  }
}

static void
set_property(GObject *object,
    guint prop_id,
    const GValue *value,
    GParamSpec *pspec)
{
  HimsdkVKBPrivate *priv;

  g_return_if_fail (IS_HIMSDK_VKB (object));
  priv = HIMSDK_VKB_GET_PRIVATE(object);

  switch (prop_id)
  {
    case PROP_KEYBOARD:
      priv->keyboard = g_value_get_object(value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
  }
}

/* Implementation of plugin interface starts here */
static void
himsdk_vkb_init (HimsdkVKB *vkb)
{
  HimsdkVKBPrivate *priv;

  g_return_if_fail (IS_HIMSDK_VKB (vkb));

  priv = HIMSDK_VKB_GET_PRIVATE (vkb);

  priv->keys          = NULL;
  priv->screen_mode   = SCREEN_NORMAL;
  priv->case_mode     = CASE_LOWER;
}

GtkWidget *
himsdk_vkb_new (HildonIMKeyboard *kbd)
{
  return g_object_new (HIMSDK_VKB_TYPE, "keyboard", kbd, NULL);
}

/* Called when the plugin is available to the user */
static void
enable (HildonIMPlugin *plugin, gboolean init)
{
  HimsdkVKBPrivate *priv;
  HimsdkVKB *vkb;

  vkb = HIMSDK_VKB (plugin);
  priv = HIMSDK_VKB_GET_PRIVATE (vkb);
  if (init == TRUE) 
  {
    hildon_im_keyboard_button_set_toggle (priv->keyboard, 
      BUTTON_MODE_A, TRUE);
    hildon_im_keyboard_button_set_toggle (priv->keyboard, 
      BUTTON_MODE_B, TRUE);
    hildon_im_keyboard_button_set_label (priv->keyboard, 
      BUTTON_MODE_A, "ABC");
    hildon_im_keyboard_button_set_label (priv->keyboard, 
      BUTTON_MODE_B, "Shift");
  }
}

/* Called when the plugin is 'disappear' */
static void
disable(HildonIMPlugin *plugin)
{
  /* empty, sic. */
}

/* Called when input mode changed */
static void
input_mode_changed (HildonIMPlugin *plugin)
{
  /* empty, sic. */
}

/* Called when the plugin is requested to 'clear'/refresh it's UI */
static void
clear(HildonIMPlugin *plugin)
{
  /* empty, sic. */
}

/* Called when the client widget changed */
static void
client_widget_changed (HildonIMPlugin *plugin)
{
  /* empty, sic. */
}

/* Called when the plugin is requested to save it's data */
static void
save_data(HildonIMPlugin *plugin)
{
  /* empty, sic. */
}

/* Called when the MODE_A button is pressed */
static void
mode_a(HildonIMPlugin *plugin)
{
  HimsdkVKBPrivate *priv;
  HimsdkVKB *vkb;

  vkb = HIMSDK_VKB (plugin);

  priv = HIMSDK_VKB_GET_PRIVATE (vkb);
  if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_B)) {
    hildon_im_keyboard_button_set_active (priv->keyboard,
        BUTTON_MODE_B, FALSE);
    if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_A)) {
      priv->case_mode = CASE_UPPER;
    } else {
      priv->case_mode = CASE_LOWER;
    }
  } else {
    if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_A)) {
      priv->case_mode = CASE_UPPER;
    } else {
      priv->case_mode = CASE_LOWER;
    }
  }
  
  update_layout (vkb);
}

/* Called when the MODE_B button is pressed */
static void
mode_b(HildonIMPlugin *plugin)
{
  HimsdkVKBPrivate *priv;
  HimsdkVKB *vkb;

  vkb = HIMSDK_VKB (plugin);

  priv = HIMSDK_VKB_GET_PRIVATE (vkb);

  if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_B)) {
    if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_A)) {
      priv->case_mode = CASE_LOWER;
    } else {
      priv->case_mode = CASE_UPPER;
    }
  } else {
    if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_A)) {
      priv->case_mode = CASE_UPPER;
    } else {
      priv->case_mode = CASE_LOWER;
    }
  }

  update_layout (vkb);
}

/* Called when the language has been changed */
static void
language (HildonIMPlugin *plugin)
{
  /* Do nothing, sic. */
  return;
}

/* CAlled when the language settings has been changed */
static void
language_settings_changed (HildonIMPlugin *plugin,
    gint index, gint value)
{
  /* empty, sic. */
}

/* Called when the backspace button is pressed */
static void
backspace (HildonIMPlugin *plugin)
{
  HimsdkVKBPrivate *priv;

  priv = HIMSDK_VKB_GET_PRIVATE (HIMSDK_VKB (plugin));
  hildon_im_keyboard_send_communication_message (priv->keyboard, 
      HILDON_IM_CONTEXT_HANDLE_BACKSPACE);
}

/* Called when the enter button is pressed */
static void
enter (HildonIMPlugin *plugin)
{
  HimsdkVKBPrivate *priv;

  priv = HIMSDK_VKB_GET_PRIVATE (HIMSDK_VKB (plugin));
  hildon_im_keyboard_send_communication_message (priv->keyboard, 
      HILDON_IM_CONTEXT_HANDLE_ENTER);
}

/* Called when the tab button is pressed */
static void
tab (HildonIMPlugin *plugin)
{
  HimsdkVKBPrivate *priv;

  priv = HIMSDK_VKB_GET_PRIVATE (HIMSDK_VKB (plugin));
  hildon_im_keyboard_send_communication_message (priv->keyboard, 
      HILDON_IM_CONTEXT_HANDLE_TAB);
}

/* Called when the standard input method settings 
 * has been changed */
static void
settings_changed (HildonIMPlugin *plugin,
    const gchar *key, const GConfValue *value)
{
  /* empty, sic. */
}

/* Supporting functions starts here */
static void 
draw_key (HimsdkVKB *vkb, VKBKey *key, GtkStateType state, gint case_mode)
{
  HimsdkVKBPrivate *priv;
  g_return_if_fail (IS_HIMSDK_VKB (vkb));
  g_return_if_fail (key != NULL);
  priv = HIMSDK_VKB_GET_PRIVATE (vkb);

  if (case_mode == CASE_UPPER) {
    pango_layout_set_text (priv->layout, key->upper, key->upper_length);
  } else { 
    pango_layout_set_text (priv->layout, key->lower, key->lower_length);
  }
  
  gtk_paint_box (priv->style, 
      GTK_WIDGET(vkb)->window, 
      state, 
      GTK_SHADOW_IN, NULL, GTK_WIDGET(vkb), NULL, 
      key->x, key->y, 
      key->width, 
      key->height);

  gtk_paint_layout (GTK_WIDGET(vkb)->style, 
      GTK_WIDGET(vkb)->window, 
      state, 
      TRUE, NULL, GTK_WIDGET(vkb), NULL, 
      key->x + KEY_LABEL_MARGIN, key->y, 
      priv->layout);
}

static void 
draw_keyboard (HimsdkVKB *vkb)
{
  HimsdkVKBPrivate *priv;
  VKBKey          *key;
  gchar           *tmp_upper,
                  *tmp_lower;

  gint i, j, x, y, width;
  gint start_x;

  g_return_if_fail (IS_HIMSDK_VKB (vkb));
  priv = HIMSDK_VKB_GET_PRIVATE (vkb);

  priv->total_keys = TOTAL_KEYPAD_KEYS;
  for (i = 0; i < KEYBOARD_ROWS; i ++)
  {
    priv->total_keys += keyboard_keys [i];
  }

  if (priv->keys == NULL)
  {
    priv->keys = g_malloc0 (sizeof(VKBKey) * priv->total_keys);
  }

  /* To handle insufficient memory error is left as an exercise
   * to the readers(TM) */

  key = priv->keys;
  
  start_x = START_X;
  y = START_Y;
  for (i = 0; i < KEYBOARD_ROWS; i ++)
  {
    x = start_x;
    tmp_upper = (gchar*) keyboard_layout [CASE_UPPER][i];
    tmp_lower = (gchar*) keyboard_layout [CASE_LOWER][i];
    for (j = 0; j < keyboard_keys [i]; j ++) 
    {
      if (i == KEYBOARD_ROWS - 1) {
        width = button_sizes [priv->screen_mode][KEYSIZE_NORMAL] * 
          keyboard_keys [1];
      }
      else if (((i % 2) == 0) &&
          ((j == 0) || 
           (j == keyboard_keys [i] - 1))
          ) {
        width = button_sizes [priv->screen_mode][KEYSIZE_BIG];
      } 
      else {
        width = button_sizes [priv->screen_mode][KEYSIZE_NORMAL];
      }
    
      key->upper  = tmp_upper;
      key->lower  = tmp_lower;
      tmp_upper   = g_utf8_find_next_char (tmp_upper, NULL);
      tmp_lower   = g_utf8_find_next_char (tmp_lower, NULL);

      if (tmp_upper != NULL) {
        key->upper_length = tmp_upper - key->upper;
      } else {
        key->upper_length = 0;
      }
      if (tmp_lower != NULL) {
        key->lower_length = tmp_lower - key->lower;
      } else {
        key->lower_length = 0;
      }
      key->x      = x;
      key->y      = y;
      key->width  = width;
      key->height = HEIGHT;
      draw_key (vkb, key, GTK_STATE_NORMAL, priv->case_mode);

      x += width;
      key ++;
    }
    y += HEIGHT;
  }

  start_x = priv->keypad_start_x =
    GTK_WIDGET (vkb)->allocation.width - 
    (button_sizes [priv->screen_mode][KEYSIZE_NORMAL] * 
    MAX_KEYPAD_KEYS + KEYPAD_MARGIN_RIGHT);
  
  y = START_Y;
  for (i = 0; i < KEYPAD_ROWS; i ++)
  {
    x = start_x;
    tmp_upper = (gchar*) keypad_layout [CASE_UPPER][i];
    tmp_lower = (gchar*) keypad_layout [CASE_LOWER][i];
    for (j = 0; j < MAX_KEYPAD_KEYS; j ++) 
    {
      width = button_sizes [priv->screen_mode][KEYSIZE_NORMAL];
      
      key->upper  = tmp_upper;
      key->lower  = tmp_lower;
      tmp_upper   = g_utf8_find_next_char (tmp_upper, NULL);
      tmp_lower   = g_utf8_find_next_char (tmp_lower, NULL);

      if (tmp_upper != NULL) {
        key->upper_length = tmp_upper - key->upper;
      } else {
        key->upper_length = 0;
      }
      if (tmp_lower != NULL) {
        key->lower_length = tmp_lower - key->lower;
      } else {
        key->lower_length = 0;
      }
      key->x      = x;
      key->y      = y;
      key->width  = width;
      key->height = HEIGHT;

      draw_key (vkb, key, GTK_STATE_NORMAL, priv->case_mode);

      x += width;
      key ++;
    }
    y += HEIGHT;
  }
}


static void 
update_layout (HimsdkVKB *vkb)
{
  HimsdkVKBPrivate *priv;

  priv = HIMSDK_VKB_GET_PRIVATE (vkb);

  draw_keyboard (vkb);
}

static void 
send_key (HimsdkVKB *vkb, VKBKey *key)
{
  HimsdkVKBPrivate *priv;
#define UTF8_SIZE 6
  gchar   text [UTF8_SIZE];

  if (key != NULL) 
  {
    priv = HIMSDK_VKB_GET_PRIVATE (vkb);    
    memset (text, 0, UTF8_SIZE);
    if (priv->case_mode == CASE_UPPER) 
    {
      memcpy (text, key->upper, key->upper_length);
    } else {
      memcpy (text, key->lower, key->lower_length);
    }

    if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_B)) {
      hildon_im_keyboard_button_set_active (priv->keyboard,
        BUTTON_MODE_B, FALSE);
      if (hildon_im_keyboard_button_get_active (priv->keyboard,
        BUTTON_MODE_A)) {
        priv->case_mode = CASE_UPPER;
      } else {
        priv->case_mode = CASE_LOWER;
      }
      update_layout (vkb);
    }
    hildon_im_keyboard_send_utf8(priv->keyboard, text);
  }
}

/* Callbacks starts here */
static gboolean 
expose_cb (GtkWidget *widget, GdkEventExpose *event)
{

  draw_keyboard (HIMSDK_VKB (widget));
  return TRUE;
}

static gboolean
destroy_cb (GtkWidget *widget, GdkEventAny *event)
{
  HimsdkVKB            *vkb;
  HimsdkVKBPrivate     *priv;

  vkb = HIMSDK_VKB (widget);
  priv = HIMSDK_VKB_GET_PRIVATE (vkb);
  if (priv->keys != NULL)
  {
    g_free (priv->keys);
    priv->keys = NULL;
  }

  return FALSE;
}

static void
realize_cb (GtkWidget *widget)
{
  HimsdkVKB            *vkb;
  HimsdkVKBPrivate     *priv;
  GdkScreen           *screen;
  GdkWindowAttr        attributes;
  gint                 attributes_mask;
  gint                 depth;

  g_return_if_fail (IS_HIMSDK_VKB (widget));
  vkb = HIMSDK_VKB (widget);
  priv = HIMSDK_VKB_GET_PRIVATE (vkb);

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
 
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = 
    gtk_widget_get_events (widget)  |
    GDK_EXPOSURE_MASK               | 
    GDK_BUTTON_PRESS_MASK           |
    GDK_BUTTON_RELEASE_MASK;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (
      gtk_widget_get_parent_window (widget), 
      &attributes, attributes_mask);

  widget->style = gtk_style_attach(widget->style, widget->window);

  gdk_window_set_user_data (widget->window, widget);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  priv->style = gtk_rc_get_style_by_paths (
      gtk_widget_get_settings(GTK_WIDGET(widget)),
      ".osso-im-key", NULL, GTK_TYPE_BUTTON);

  screen = gtk_widget_get_screen(widget);
  depth = DefaultDepth (
      GDK_SCREEN_XDISPLAY (screen), 
      GDK_SCREEN_XNUMBER (screen));

  priv->layout = pango_layout_new (
      gtk_widget_get_pango_context (widget));
}

static void 
size_request_cb (GtkWidget *widget, 
    GtkRequisition *requisition)
{
}

static void 
size_allocate_cb (GtkWidget *widget, 
    GtkAllocation *allocation)
{
  HimsdkVKBPrivate     *priv;

  g_return_if_fail (IS_HIMSDK_VKB (widget));
  priv = HIMSDK_VKB_GET_PRIVATE (HIMSDK_VKB (widget));

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
  {
    gdk_window_move_resize (widget->window, 
        allocation->x, allocation->y,
        allocation->width, allocation->height);

    if (allocation->width > FULLSCREEN_WIDTH) {
      priv->screen_mode = SCREEN_FULL;
    } else {
      priv->screen_mode = SCREEN_NORMAL;
    }
    draw_keyboard (HIMSDK_VKB (widget));
  }
}

static VKBKey *
get_key (HimsdkVKB *vkb, gint x, gint y)
{
  HimsdkVKBPrivate   *priv;

  gint row;
  gint keyboard_width;
  gint i, offset, size;

  priv = HIMSDK_VKB_GET_PRIVATE (vkb);
  
  keyboard_width = button_sizes [priv->screen_mode][KEYSIZE_NORMAL] * 
    keyboard_keys [1] + START_X;

  /* Ignore if it fall on the 'dead zones' */
  if ((x < START_X) ||
      (y < START_Y) ||
      ((x > keyboard_width) && 
       (x < priv->keypad_start_x)) ||
      (x > GTK_WIDGET (vkb)->allocation.width - KEYPAD_MARGIN_RIGHT)
      ) {    
    return NULL;
  }

  row = y / HEIGHT;
  if ((x < keyboard_width) &&
          (row == (KEYBOARD_ROWS - 1)))
  {
    return priv->keys + (priv->total_keys - TOTAL_KEYPAD_KEYS - 1);
  }

  if (x > priv->keypad_start_x) 
  {
    i = 0;
    while (i < MAX_KEYPAD_KEYS)
    {
      if (x > (priv->keypad_start_x + 
            button_sizes [priv->screen_mode][KEYSIZE_NORMAL] * 
            (i + 1)))
      {
        i ++;
      } else {
        break;
      }
    }
    return priv->keys + (priv->total_keys - 
        TOTAL_KEYPAD_KEYS + 
        (MAX_KEYPAD_KEYS * row) +
        i
        );
  } else {
    if (row == 1) {
      i = 0;
      while (i < keyboard_keys [1])
      {
        if (x > (START_X + 
              button_sizes [priv->screen_mode][KEYSIZE_NORMAL] * 
              (i + 1)))
        {
          i ++;
        } else {
          break;
        }
      }
      return priv->keys + (keyboard_keys [0] + i);
    } else {
      offset = 0;
      for (i = 0; i < row; i ++) 
      {
        offset += keyboard_keys [i];
      }

      size = button_sizes [priv->screen_mode][KEYSIZE_BIG];
      if ((x < size + START_X)) {
        i = 0;
      } else if (x > keyboard_width - size)         
      {
        i = keyboard_keys [row] - 1;
      } else {
        i = 0;
        while (i < keyboard_keys [row])
        {
          if (x > (START_X + size +
                button_sizes [priv->screen_mode][KEYSIZE_NORMAL] * 
                (i + 1)))
          {
            i ++;
          } else {
            break;
          }
        }
        i ++;
      }
      return priv->keys + (offset + i);
    }
  }
  return NULL;
}

static gboolean
button_press_cb (GtkWidget *widget, 
    GdkEventButton *event)
{
  VKBKey *key;
  HimsdkVKB *vkb;
  HimsdkVKBPrivate *priv;

  vkb = HIMSDK_VKB (widget);
  key = get_key (vkb, event->x, event->y);

  if (key != NULL) 
  {
    priv = HIMSDK_VKB_GET_PRIVATE (vkb);
    draw_key (vkb, key, GTK_STATE_ACTIVE, priv->case_mode);
    priv->last_key = key;
    
  }
  return FALSE;
}

static gboolean
button_release_cb (GtkWidget *widget, 
    GdkEventButton *event)
{
  HimsdkVKB *vkb;
  HimsdkVKBPrivate *priv;

  vkb = HIMSDK_VKB (widget);
  priv = HIMSDK_VKB_GET_PRIVATE (vkb);

  if (priv->last_key != NULL) 
  {
    send_key (vkb, priv->last_key);
    draw_key (HIMSDK_VKB (widget), 
        priv->last_key, 
        GTK_STATE_NORMAL,
        priv->case_mode);
    priv->last_key = NULL;
  }
  return FALSE;
}

/* The end of the road */
